/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.platform.raspberry.generator.extension.brick;

import org.fortiss.af3.platform.raspberry.model.brick.AccelerationXSensor;
import org.fortiss.af3.platform.raspberry.model.brick.UIDUnit;

/**
 * Base class for {@link AccelerationXSensor} and similar sensors generator extensions.
 * 
 * @author hoelzl
 */
abstract class IMUSensorGeneratorExtensionBase<T extends UIDUnit> extends
		UIDUnitSharedInstanceBrickGeneratorExtensionBase<T> {
	/** Constructor. */
	public IMUSensorGeneratorExtensionBase(T modelElement) {
		super(modelElement);
	}

	/** {@inheritDoc} */
	@Override
	public String getHeaderFileName() {
		return "brick_imu_v2.h";
	}

	/** {@inheritDoc} */
	@Override
	public String getSharedInstanceVariableDeclarationCode() {
		String postfix = getSharedInstancePostfix();
		StringBuilder sb = new StringBuilder();
		sb.append("static float Q_" + postfix + " = 0.022;\n");
		sb.append("static float R_" + postfix + " = 0.917;\n");
		sb.append("static float estimates_" + postfix + "[6] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0};\n");
		sb.append("static float last_p_" + postfix + "[6] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0};\n");
		sb.append("static float acceleration_X_" + postfix + ", acceleration_Y_" + postfix +
				", acceleration_Z_" + postfix + ";\n");
		sb.append("static uint64_t acc_last_cb_time_" + postfix + " = 0;\n");
		sb.append("static float angularVelocity_X_" + postfix + ", angularVelocity_Y_" + postfix +
				", angularVelocity_Z_" + postfix + ";\n");
		sb.append("static uint64_t angV_last_cb_time_" + postfix + " = 0;\n");
		sb.append("static IMUV2 imu_v2_device_" + postfix + ";\n");
		return sb.toString();
	}

	/** {@inheritDoc} */
	@Override
	public String getSharedInstanceAuxiliaryFunctions() {
		String postfix = getSharedInstancePostfix();
		StringBuilder sb = new StringBuilder();
		// kalman filter
		sb.append("static float kalman_filter_" + postfix + "(float value, int sensor) {\n");
		sb.append("if(sensor < 0 || sensor > 5) return -1;\n");
		sb.append("float temp_est = estimates_" + postfix + "[sensor];\n");
		sb.append("float p_temp = last_p_" + postfix + "[sensor] + Q_" + postfix + ";\n");
		sb.append("float K = p_temp * (1.0/(p_temp + R_" + postfix + "));\n");
		sb.append("float est = temp_est + K * (value - temp_est);\n");
		sb.append("float P = (1-K)*p_temp;\n");
		sb.append("last_p_" + postfix + "[sensor] = P;\n");
		sb.append("estimates_" + postfix + "[sensor] = est;\n");
		sb.append("return est;\n");
		sb.append("}\n\n");
		// acc callback
		sb.append("static void acc_callback_" + postfix +
				"(int16_t x, int16_t y, int16_t z, void* data) {\n");
		sb.append("if (abs(x - acceleration_X_" + postfix + ") > 2 * 10 * 100 ||\n");
		sb.append("abs(y - acceleration_Y_" + postfix + ") > 2 * 10 * 100 ||\n");
		sb.append("abs(z - acceleration_Z_" + postfix + ") > 2 * 10 * 100)\n");
		sb.append("    return;\n");
		sb.append("acceleration_X_" + postfix + " = kalman_filter_" + postfix + "(x, 0) / 100.0;\n");
		sb.append("acceleration_Y_" + postfix + " = kalman_filter_" + postfix + "(y, 1) / 100.0;\n");
		sb.append("acceleration_Z_" + postfix + " = kalman_filter_" + postfix + "(z, 2) / 100.0;\n");
		sb.append("acc_last_cb_time_" + postfix + " = time_util_get_current_micros();\n");
		sb.append("}\n\n");
		// acc validity function
		sb.append("static inline bool acc_is_invalid_" + postfix + "() {\n");
		sb.append("\treturn time_util_get_elapsed_micros_since(acc_last_cb_time_");
		sb.append(postfix);
		sb.append(") >= ((uint64_t)1 * SECONDS_IN_MICROS);\n");
		sb.append("}\n\n");
		// angle callback
		sb.append("static void angV_callback_" + postfix +
				"(int16_t x, int16_t y, int16_t z, void* data) {\n");
		sb.append("angularVelocity_X_" + postfix + " = kalman_filter_" + postfix + "(x, 3) / 16;\n");
		sb.append("angularVelocity_Y_" + postfix + " = kalman_filter_" + postfix + "(y, 4) / 16;\n");
		sb.append("angularVelocity_Z_" + postfix + " = kalman_filter_" + postfix + "(z, 5) / 16;\n");
		sb.append("angV_last_cb_time_" + postfix + " = time_util_get_current_micros();\n");
		sb.append("}\n\n");
		// angle validity function
		sb.append("static inline bool angV_is_invalid_" + postfix + "() {\n");
		sb.append("\treturn time_util_get_elapsed_micros_since(angV_last_cb_time_");
		sb.append(postfix);
		sb.append(") >= ((uint64_t)1 * SECONDS_IN_MICROS);\n");
		sb.append("}\n\n");
		return sb.toString();
	}

	/** Returns the name for the acceleration value variable. */
	protected final String getAccValueVariable(String axis) {
		return "acceleration_" + axis + "_" + getSharedInstancePostfix();
	}

	/** Returns the name for the angular velocity value variable. */
	protected final String getAngVValueVariable(String axis) {
		return "angularVelocity_" + axis + "_" + getSharedInstancePostfix();
	}

	/** Returns the function name for the acceleration invalidity check. */
	protected final String getAccInvalidCheckFunctionCall() {
		return "acc_is_invalid_" + getSharedInstancePostfix() + "()";
	}

	/** Returns the function name for the angular velocity invalidity check. */
	protected final String getAngVInvalidCheckFunctionCall() {
		return "angV_is_invalid_" + getSharedInstancePostfix() + "()";
	}

	/** {@inheritDoc} */
	@Override
	public String getSharedInstanceInitializationCode() {
		String shared = getSharedInstancePostfix();
		StringBuilder sb = new StringBuilder();
		UIDUnit sensor = modelElement;
		sb.append("// Initialize IMU @ " + sensor.getUniqueBrickletID() + "\n");
		sb.append("imu_v2_create(&imu_v2_device_" + shared + ", \"" + sensor.getUniqueBrickletID() +
				"\", &" + getConnectionVariable() + ");\n");
		sb.append("imu_v2_register_callback(&imu_v2_device_" + shared +
				", IMU_V2_CALLBACK_LINEAR_ACCELERATION, (void*)acc_callback_" + shared +
				", NULL);\n");
		sb.append("imu_v2_set_linear_acceleration_period(&imu_v2_device_" + shared + ", 5);\n");
		sb.append("imu_v2_register_callback(&imu_v2_device_" + shared +
				", IMU_V2_CALLBACK_ANGULAR_VELOCITY, (void*)angV_callback_" + shared + ", NULL);\n");
		sb.append("imu_v2_set_angular_velocity_period(&imu_v2_device_" + shared + ", 5);\n");
		return sb.toString();
	}

	/** {@inheritDoc} */
	@Override
	public String getSharedInstanceReadCode() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public String getSharedInstanceWriteCode() {
		return null;
	}
}
